#define _LARGEFILE64_SOURCE
#define _FILE_OFFSET_BITS 64

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <fcntl.h>
#include <assert.h>

#include <scsi/sg.h>
#include <scsi/scsi.h>

#include "sgilabel.h"
#include "dk.h"

#define SENSE_BUFF_LEN 32       /* Arbitrary, could be larger */
#define MODE_SENSE6_CMD      0x1a
#define MODE_SENSE6_CMDLEN   6

#define PG_CODE_ALL 0x3f
#define PG_CODE_MASK 0x3f
#define PG_CODE_MAX 0x3f
#define SPG_CODE_ALL 0xff

#define DEF_TIMEOUT 60000

void dump_hex(uint8_t *buf, int size) {
	int		i;
	unsigned char	ch;
	char		sascii[17];

	sascii[16]=0x0;

	for(i=0;i<size;i++) {
		ch=buf[i];
		if (i%16 == 0) {
			fprintf(stderr, "%04x ", i);
		}
		fprintf(stderr, "%02x ", ch);
		if (ch >= ' ' && ch <= '}')
			sascii[i%16]=ch;
		else
			sascii[i%16]='.';

		if (i%16 == 15)
			fprintf(stderr, "   %s\n", sascii);
	}

	/* i++ after loop */
	if (i%16 != 0) {

		for(;i%16 != 0;i++) {
			fprintf(stderr, "   ");
			sascii[i%16]=' ';
		}

		fprintf(stderr, "   %s\n", sascii);
	}
	fprintf(stderr, "\n");
}

static int do_modes(int sg_fd, int pc, int pg_code, int sub_pg_code,
                    void *resp, int resplen)
{
    int res, k;
    unsigned char modesCmdBlk[MODE_SENSE6_CMDLEN] = 
        {MODE_SENSE6_CMD, 0, 0, 0, 0, 0};
    unsigned char sense_b[SENSE_BUFF_LEN];
    struct sg_io_hdr io_hdr;
    int cmd_len;

    modesCmdBlk[1] = (unsigned char)(0);
    modesCmdBlk[2] = (unsigned char)(((pc << 6) & 0xc0) | 
                                     (pg_code & PG_CODE_MASK));
    modesCmdBlk[3] = (unsigned char)(sub_pg_code & 0xff);
    modesCmdBlk[4] = (unsigned char)(resplen & 0xff);
#if 0
    fprintf(stderr, "        mode sense cdb: ");
    for (k = 0; k < MODE_SENSE6_CMDLEN; ++k)
	fprintf(stderr, "%02x ", modesCmdBlk[k]);
    fprintf(stderr, "\n");
#endif
    memset(&io_hdr, 0, sizeof(struct sg_io_hdr));
    memset(sense_b, 0, sizeof(sense_b));

    io_hdr.interface_id = 'S';
    io_hdr.cmd_len = MODE_SENSE6_CMDLEN;
    io_hdr.mx_sb_len = sizeof(sense_b);
    io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
    io_hdr.dxfer_len = resplen;
    io_hdr.dxferp = resp;
    io_hdr.cmdp = modesCmdBlk;
    io_hdr.sbp = sense_b;
    io_hdr.timeout = DEF_TIMEOUT;

    if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
        perror("SG_IO (mode sense) error");
        return -1;
    }

    if (io_hdr.status != 0 || io_hdr.host_status != 0 || io_hdr.driver_status != 0)
    	return -1;

    return 0;
}



#define SENSE_BUFF_SZ 32
#define RCAP_REPLY_LEN 8

/* Performs a 10 byte READ CAPACITY command and fetches response.
 * Return of 0 -> success, -1 -> failure */
int do_readcap_10(int sg_fd, int pmi, unsigned int lba, 
                  unsigned int * last_sect, unsigned int * sect_sz)
{
	int res;
	unsigned char rcCmdBlk[10] = {READ_CAPACITY, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	unsigned char rcBuff[RCAP_REPLY_LEN];
	unsigned char sense_b[SENSE_BUFF_SZ];
	struct sg_io_hdr io_hdr;

	if (pmi) { /* lbs only valid when pmi set */
		rcCmdBlk[8] |= 1;
		rcCmdBlk[2] = (lba >> 24) & 0xff;
		rcCmdBlk[3] = (lba >> 16) & 0xff;
		rcCmdBlk[4] = (lba >> 8) & 0xff;
		rcCmdBlk[5] = lba & 0xff;
	}

	memset(&io_hdr, 0, sizeof(struct sg_io_hdr));

	io_hdr.interface_id = 'S';
	io_hdr.cmd_len = sizeof(rcCmdBlk);
	io_hdr.mx_sb_len = sizeof(sense_b);
	io_hdr.dxfer_direction = SG_DXFER_FROM_DEV;
	io_hdr.dxfer_len = sizeof(rcBuff);
	io_hdr.dxferp = rcBuff;
	io_hdr.cmdp = rcCmdBlk;
	io_hdr.sbp = sense_b;
	io_hdr.timeout = 60000;

	if (ioctl(sg_fd, SG_IO, &io_hdr) < 0) {
		perror("read_capacity (SG_IO) error");
		return -1;
	}

	if (io_hdr.status != 0 || io_hdr.host_status != 0 || io_hdr.driver_status != 0)
		return -1;

	*last_sect = ((rcBuff[0] << 24) | (rcBuff[1] << 16) |
		(rcBuff[2] << 8) | rcBuff[3]);

	*sect_sz = (rcBuff[4] << 24) | (rcBuff[5] << 16) |
		(rcBuff[6] << 8) | rcBuff[7];

	return 0;
}

struct sgpage_header {
	uint8_t		length;
	uint8_t		medium_type;
	uint8_t		devspecific;
	uint8_t		bdlen;
};

struct sgpage_blockdesc {
	uint8_t		nblocks[4];
	uint8_t		reserved;
	uint8_t		blocklen[3];
};

struct sgpage_geometry {
	uint8_t		pagecode;
	uint8_t		pagelen;
	uint8_t		cylinders[3];
	uint8_t		heads;
	uint8_t		startcylwriteprecomp[3];
	uint8_t		startcylredwritecurrent[3];
	uint8_t		devsteprate[2];
	uint8_t		landzonecylinder[3];
	uint8_t		reserved1;
	uint8_t		rotoffset;
	uint8_t		reserver2;
	uint8_t		mediumrotationrate[2];
	uint8_t		reserved3[2];
};

struct sgpage_format {
	uint8_t			pagecode;
	uint8_t			pagelen;
	uint8_t			tracksperzone[2];
	uint8_t			alt_sectors_per_zone[2];
	uint8_t			alt_tracks_per_zone[2];
	uint8_t			alt_tracks_per_lun[2];
	uint8_t			sectors_per_track[2];
	uint8_t			bytes_per_sector[2];
	uint8_t			interleave[2];
	uint8_t			trackskew[2];
	uint8_t			cylinderskew[2];
	uint8_t			reserved[3];
};

#define MAX_RESPONSE	252

struct diskinfo {
	unsigned int	sectsize;
	unsigned int	lastsect;
	unsigned int	heads;
	unsigned int	cylinders;
	unsigned int	sectors;

	unsigned int	numblocks; /* Block Descriptor - most likely 0 */
};

unsigned int getbigendian(uint8_t *val, int bytes) {
	unsigned int	v=0;
	int		i;
	for(i=bytes;i>0;i--,v<<8,v|=*val++);
	return v;
}

struct diskinfo *getdiskinfo(int fd) {
	int			len, rc;
	unsigned int		sectorspt, heads, cylinders, numblocks;
	uint8_t			response[MAX_RESPONSE];
	struct sgpage_format	*fm;
	struct sgpage_blockdesc	*bd;
	struct sgpage_geometry	*geom;
	struct sgpage_header	*head;
	static struct diskinfo	di;

	/* Read capacity e.g. sector size & last sector */
	if (do_readcap_10(fd, 0, 0, &di.lastsect, &di.sectsize) != 0) {
		fprintf(stderr, "SCSI Read capacity failed\n");
		return NULL;
	}

	/* Page Code = 0, Page = 3, SubPage = 0 */
	if(do_modes(fd, 0, 3, 0, &response, MAX_RESPONSE)) {
		fprintf(stderr, "SCSI Get ModePage 3 failed\n");
		return NULL;
	}

	head=(struct sgpage_header *) &response[0];
	fm=(struct sgpage_format *) &response[head->bdlen+sizeof(struct sgpage_header)];
	bd=(struct sgpage_blockdesc *) &response[sizeof(struct sgpage_header)];

	di.numblocks=getbigendian(bd->nblocks, 4);
	di.sectors=getbigendian(fm->sectors_per_track, 2);

	if (do_modes(fd, 0, 4, 0, &response, MAX_RESPONSE)) {
		fprintf(stderr, "SCSI Get ModePage 4 failed\n");
		return NULL;
	}

	head=(struct sgpage_header *) &response[0];
	geom=(struct sgpage_geometry *) &response[head->bdlen+sizeof(struct sgpage_header)];

	di.heads=geom->heads;
	di.cylinders=getbigendian(geom->cylinders, 3);

	return &di;
}

struct partinfo {
	unsigned int	heads;
	unsigned int	sectors;
	struct {
		unsigned int start;
		unsigned int size;
	} partition[16];
};

struct partinfo *readvolhdr(int fd) {
	int			i;
	uint8_t			vhdr[512];
	sgi_partition		*part;
	static struct partinfo	pi;

	memset(&pi, 0, sizeof(struct partinfo));

	lseek(fd, 0, SEEK_SET);
	i=read(fd, &vhdr, 512);

	part=(sgi_partition *) &vhdr;

	if (part->magic != SGI_LABEL_MAGIC) {
		fprintf(stderr, "SGI Disklabel magic mismatch\n");
		return NULL;
	}

	pi.sectors=SSWAP16(part->devparam.nsect);
	pi.heads=SSWAP16(part->devparam.ntrks);

	for(i=0;i<16;i++) {
		pi.partition[i].size=SSWAP32(part->partitions[i].num_sectors);
		pi.partition[i].start=SSWAP32(part->partitions[i].start_sector);

		printf("Part %2d Start %8d Size %8d\n", i,
			pi.partition[i].start, pi.partition[i].size);
	}

	return &pi;
};

int writeblock(int fd, off64_t off, void *buffer, int size) {
	off64_t		noff;
	size_t		len;

	noff=lseek64(fd, off, SEEK_SET);

	if (noff != off) {
		fprintf(stderr, "lseek64 failed\n");
		return 0;
	}

	fprintf(stderr, "seeked to %llu\n", noff);

	len=write(fd, buffer, size);

	if (len < size) {
		fprintf(stderr, "write of index block failed\n");	
		return 0;
	}

	return 1;
}

int writesnitrail(int fd, struct diskinfo *di, struct partinfo *pi) {
	struct dkindex_ondisk	dki;
	struct dkpart_ondisk	dkp;
	off64_t			off;
	off64_t			block1, block2;
	int			i;
	int			len;
	int			secpercyl;

	unsigned int secspercyl;
	unsigned int realcylinders;
	unsigned int realsize;

	memset(&dki, 0x0, sizeof(struct dkindex_ondisk));

	dki.dk_version=SSWAP32(1);
	dki.dk_magic=SSWAP32(DK_XMAGIC);

	for(i=0;i<DK_NCOPIES;i++) {
		dki.dkx_v1.dkx_asm[i].cylinder=0xffffffff;
		dki.dkx_v1.dkx_type[i].cylinder=0xffffffff;
		dki.dkx_v1.dkx_part[i].cylinder=0xffffffff;
		dki.dkx_v1.dkx_tune[i].cylinder=0xffffffff;
		dki.dkx_v1.dkx_misc[i].cylinder=0xffffffff;
	}

	secspercyl=di->sectors*di->heads;
	realcylinders=di->lastsect/secspercyl;
	realsize=realcylinders*secspercyl;

	block1=realsize-(1*di->sectors);
	block2=realsize-(2*di->sectors);

	printf("Block1: %lu Block2: %lu\n",
		realsize-(2*di->sectors),
		realsize-(1*di->sectors));

	for(i=0;i<DK_NCOPIES;i++) {
		dki.dkx_v1.dkx_part[i].sector=block1+4;
		dki.dkx_v1.dkx_part[i].cylinder=0;
	}

	dki.dk_level=SSWAP32(1);

	if (!writeblock(fd, block1*512, &dki, 2048))
		return 0;

	if (!writeblock(fd, block2*512, &dki, 2048))
		return 0;

	memset(&dkp, 0, sizeof(struct dkpart_ondisk));
	dkp.dk_version=SSWAP32(1);
	dkp.dk_magic=SSWAP32(DK_MAGIC);

	secpercyl=pi->heads*pi->sectors;

	dkp.dkp_v1.dkp_pcylinders=di->lastsect/secpercyl;
	dkp.dkp_v1.dkp_ucylinders=di->lastsect/secpercyl-2;
	dkp.dkp_v1.dkp_heads=pi->heads;
	dkp.dkp_v1.dkp_sectors=pi->sectors;
	dkp.dkp_v1.dkp_cylsize=secpercyl;

	printf("Converting partition table:\n");

	for(i=0;i<16;i++) {

		printf("\tPartition %2d Offset %8d Length %8d\n",
				i,
				pi->partition[i].start/secpercyl,
				pi->partition[i].size/secpercyl);
		printf("\t\tStart: %8d Size: %8d\n", pi->partition[i].start, pi->partition[i].size);

		if (pi->partition[i].start % secpercyl != 0) {
			fprintf(stderr, "\tPartition %d start not on cylinders boundary\n", i);
		}

		if (pi->partition[i].size % secpercyl != 0) {
			fprintf(stderr, "\tPartition %d size not on cylinders boundary\n", i);
		}

		/* Cylinders */
		dkp.dkp_v1.dkp_partition[i].dkp_offset=
				pi->partition[i].start/secpercyl;

		dkp.dkp_v1.dkp_partition[i].dkp_length=
				pi->partition[i].size/secpercyl;
	}

	dkp.dk_level=SSWAP32(1);

	if (!writeblock(fd, (block1+4)*512, &dkp, 2048))
		return 0;

	return 1;
}

int main(int argc, char **argv) {
	int			fd;
	struct diskinfo		*di;
	struct partinfo		*pi;


	fd=open("/dev/sda", O_RDWR | O_NONBLOCK);
	assert(fd>=0);

	di=getdiskinfo(fd);

	if (!di) {
		close(fd);
		exit(-1);
	}
	pi=readvolhdr(fd);
	if (!pi) {
		close(fd);
		exit(-1);
	}

	writesnitrail(fd, di, pi);

	close(fd);
}

